{-

SPS/APS - Security Protocol Specification langauge
Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2013-2015

All Rights Reserved.

-}
module ComposeM where
import System.Environment
import Data.List
import Preprocessor
import Msg
-- import Debug.Trace
-------------------------------------------------------------
-- composable k m | trace ("composable\n\tterm: " ++ show m ++ "\n\tknowledge:" ++ show k) False = undefined
composable k m@(Atom x)= compose k m /= []
-------------------------------------------------------------------
-- g^(xy) = g(yx)
composable k m@(Comp Exp opt1 [Comp Exp opt2 [g, x],y])= ((compose k m) /= []) || 
                                                ((composable k (Comp Exp opt1 [g, x])) && (composable k y)) ||
                                                ((composable k (Comp Exp opt1 [g, y])) && (composable k x)) 
-- (a.b.c) = a.b.c = (a.b).c = (a.c).b = (b.c).a
composable k m@(Comp Mult opt1 [Comp Mult opt2 [x,y],z])= ((compose k m) /= []) || 
                                                    ((composable k x) && (composable k y) && (composable k z)) ||
                                                    (composable k (Comp Mult opt1  [x,y]) && (composable k z)) ||
                                                    (composable k (Comp Mult opt1 [x,z]) && (composable k y)) ||
                                                    (composable k (Comp Mult opt1 [y,z]) && (composable k x))
-- REST MUST BE HANDLED WITH NEXT CASE
-- OR approxF                                        -- (composable k (Comp Mult [Comp Mult [y,x],z])) ||
                                                    -- (composable k (Comp Mult [Comp Mult [z,x],y])) || 
                                                    -- (composable k (Comp Mult [Comp Mult [x,z],y])) || 
                                                    -- (composable k (Comp Mult [Comp Mult [z,y],x])) || 
                                                    -- (composable k (Comp Mult [Comp Mult [y,z],x])) 
-- a.b = b.a
composable k m@(Comp Mult opt [x,y])= ((compose k m) /= []) || ((composable k x) && (composable k y)) 
-- this is covered by approxF                        || (composable k (Comp Mult [y,x])) 
-------------------------------------------------------------------
composable k m@(Comp Apply opt1 lst) = (compose k m /= []) || ((and (map (composable k) (tail lst))) && (not (isMap (head lst)))) --HERE
composable k m@(Comp op opt1 lst)    =     if (compose k m /= []) then True
                                                else if (not(isPrivate op)) then and (map (composable k) lst)
                                                else False
---------------------------------------------------------------
notComp k m= not (composable k m)
-----------------------------------------------------------------
composeErr k m= if  ((compose k m) == []) then error (show(m) ++" is NOT composable"++ show(k)) else head (compose k m)
--------------------------------------------------
compose k m= [x | (x,y)<-k, approxF y m]
-----------------------------------------------------------------------------------
approxF m1@(Comp Mult opt1 [a1, Comp Mult opt2 [b1,c1]]) m2@(Comp Mult opt3 [a2, Comp Mult opt4 [b2,c2]]) = 
                                                        ((approxF a1 a2) &&  (approxF b1 b2) && (approxF c1 c2)) ||
                                                        ((approxF a1 a2) &&  (approxF b1 c2) && (approxF c1 b2)) ||
                                                        ((approxF a1 b2) &&  (approxF b1 a2) && (approxF c1 c2)) ||
                                                        ((approxF a1 b2) &&  (approxF b1 c2) && (approxF c1 a2)) ||
                                                        ((approxF a1 c2) &&  (approxF b1 b2) && (approxF c1 a2)) ||
                                                        ((approxF a1 c2) &&  (approxF b1 a2) && (approxF c1 b2))
approxF m1@(Comp Mult opt1 [a,b]) m2@(Comp Mult opt2 [c,d])= (approxF a c && approxF b d) || (approxF a c &&approxF b d)
approxF m1@(Comp Exp opt1[(Comp Exp opt2[g1,x1]),y1]) m2@(Comp Exp opt3 [(Comp Exp opt4 [g2,x2]),y2])= approxF g1 g2 && ((approxF x1 x2 && approxF y1 y2) || (approxF x1 y2 && approxF y1 x2))
approxF m1 m2= m1==m2 -- or real work!
-----------------------------------------------------------------
--approxfLst is approxf for a pair of lists of  3-elements each:
--- e.g., (["X","Y","Z"],["A","B","C"])
approxPrLst [a,b,c][x,y,z]= (approxF a x) && (approxF b y) && (approxF c z)
approxfPer x = or [ (approxPrLst a b) | (a,b)<-x ]
-----------------------------
isPrivate op = op `elem` [Pk, Inv, Shk, Sk]
isMap (Atom ('m':'p':'p':rst)) = True
isMap  _ = False

composE k m@(Atom x) = composeErr k m
composE k m@(Comp Exp opt1 [Comp Exp opt2 [g, x],y])= if ((compose k m) /= []) then head (compose k m) 
                                             else if ((composable k (Comp Exp opt1 [g, x])) && (composable k y)) 
                                                then Comp Exp opt1 [composE k (Comp Exp opt2 [g, x]), composE k y] --
                                             else if ((composable k (Comp Exp opt1 [g, y])) && (composable k x)) 
                                                then Comp Exp opt1 [composE k (Comp Exp opt2 [g, y]), composE k x] --
                                             else error (show(m) ++" is NOT composable;; from\n"++show(k)) -- here is the bad news
composE k m@(Comp Mult opt1 [Comp Mult opt2 [x,y],z])= if ((compose k m) /= []) then head (compose k m)
                                        else if ((composable k x) && (composable k y) && (composable k z)) 
                                         then Comp Mult opt1 [Comp Mult opt2 [(composE k y),(composE k y)],(composE k z)]
                                        else if (composable k (Comp Mult opt1 [y,x]) && (composable k z)) 
                                         then Comp Mult opt1 [(composE k (Comp Mult opt2 [y,x])),(composE k z)] 
                                        else if (composable k (Comp Mult opt1 [x,y]) && (composable k z)) --
                                         then Comp Mult opt1 [(composE k (Comp Mult opt2 [x,y])),(composE k z)]
                                        else if (composable k (Comp Mult opt1 [x,z]) && (composable k y))
                                         then Comp Mult opt1 [(composE k (Comp Mult opt2 [x,z])),(composE k y)]                        
                                        else if (composable k (Comp Mult opt1 [z,x]) && (composable k y)) --
                                         then Comp Mult opt1 [(composE k (Comp Mult opt2 [z,x])),(composE k y)]
                                        else if (composable k (Comp Mult opt1 [y,z]) && (composable k x))
                                         then Comp Mult opt1 [(composE k (Comp Mult opt2 [y,z])),(composE k x)]                        
                                        else if (composable k (Comp Mult opt1 [z,y]) && (composable k x)) --
                                          then Comp Mult opt1 [(composE k (Comp Mult opt2 [z,y])),(composE k x)]
                                        else error (show(m) ++" is NOT composable;;; from\n"++show(k)) -- here is the bad news
composE k m@(Comp Mult opt1 [x,y])= if ((compose k m) /= []) then head (compose k m)
                                else if ((composable k x) && (composable k y)) then Comp Mult opt1 [composE k x, composE k y]
                                else if (composable k (Comp Mult opt1 [y,x])) then head (compose k (Comp Mult opt1 [y,x]))
                                else error (show(m) ++" is NOT composable;, from\n"++show(k)) -- here is the bad news
composE k m@(Comp Apply opt lst) = if (compose k m /= []) then head (compose k m) 
                                else Comp Apply (composeAnnot k opt) ((head lst):(map (composE k) (tail lst)))
composE k m@(Comp op opt lst)=     if  ((compose k m) /= []) then head (compose k m) 
                            else if (not(isPrivate op)) then Comp op (composeAnnot k opt) (map (composE k) lst)
                            else error (show(m) ++" is NOT composable;,; from\n"++show(k)) -- here is the bad news
----------------------------------------------------------------------------------
-- composeAnnot k opt=opt
composeAnnot  k (FunSign _:rest) = composeAnnot k rest
composeAnnot  k (Option ((at,Atom val)):rest) = if( (isNum val) || (isConst val)) then Option ((at, Atom val)): composeAnnot k rest
                                                    else Option ((at,(composE k (Atom val)))): composeAnnot k rest
composeAnnot k [] = []
isNum x= elem (head x) ['1' .. '9']
isConst x = (head x) == 'c'
-------------
composeType k t= (getType t) (composeAnnot k (getAnnot t))
-------------------------------------------------------------------
composeDet k (ch@((p1Id,p1M,p1b,p1Ps),chtype,(p2Id,p2M,p2b,p2Ps)),m,mm1,mm2) =  (((p1Id,p1M',p1b,p1Ps),chtype,(p2Id,p2M',p2b,p2Ps)),m,mm1,mm2) 
                                                                                where p1M' = composE k p1M;
                                                                                      p2M' = composE k p2M;
