{

{-
SPS/APS - Security Protocol Specification langauge
Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2013-2015

APCC - Automated Protocol Composition Checker

Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2014-2015
Copyright Paolo Modesti, School of Computing Science, Newcastle University 2015
 
Open Source Fixedpoint Model-Checker version 2012b

(C) Copyright Paolo Modesti 2012
(C) Copyright Sebastian Moedersheim 2010,2012
(C) Copyright IBM Corp. 2009

All Rights Reserved.

-}

module Parser where
import Lexer
import Ast
import Msg
import Data.List
import Control.Monad

-- signOrCrypt k@(Comp Inv x) lst = Comp Sign [k,lst] -- Working 12.6
-- signOrCrypt k lst = Comp Crypt [k,lst] -- Working 12.6
 --signOrCrypt k@(Comp Inv opt x) lst = Comp Sign [] (k:lst)
-- signOrCrypt k lst = Comp Crypt [] (k:lst)
-- atomizeOptions: Deals with the annotations on Types and Operators to ALLOW users to "negotiate" attributes
-- atomizeOpts []=[]
-- atomizeOpts (at,val):rest=(at, Atom val): atomizeOpts rest

}


%name apspareser
%tokentype {Token}

%token        
  ident                { TATOM _ $$}
  "Abstraction"        { TABSTRACTION _}
  "Actions"            { TACTIONS _}
  "Agents"             { TAGENTS _}
  "as"            { TAS _}
  "*->"            { TAUTHCH _ }
  "authenticates"    { TAUTHENTICATES _}
  "between"     { TBETWEEN _}
  "}"    { TCLOSEB _ }
  ")"    { TCLOSEP _ }
  "|}"       { TCLOSESCRYPT _}
  "]"    { TCLOSESQB _ }
  ":"    { TCOLON _  }
  ","    { TCOMMA _ }
  "->*"    { TCONFCH _ }
  "."    { TDOT _ }
  "!"     { TEXCLAM  _}
  "*->>"    { TFAUTHCH _}
  "fi"     { TFI _}
  "*->>*"    { TFSECCH _}
  "Goals"    { TGOALS _}
  "if"     { TIF _}
  "->"    { TINSECCH _}
  "Knowledge"    { TKNOWLEDGE _}
  "{"         { TOPENB _  }
  "("    { TOPENP _  }
  "{#"                        {TBEGINCOMMENT _}
  "#}"                        {TENDCOMMENT _}
  "{|"         { TOPENSCRYPT _}
  "["    { TOPENSQB _ }
  "%"    { TPERCENT _}
  "Protocol"    { TPROTOCOL _ }
  "?"     { TQUESTION _}
  "*->*"        { TSECCH _  }
  "secret"     { TSECRET _}
  ";"           { TSEMICOLON _ }
  "then"     { TTHEN _}
  "Types"    { TTYPES _}
  "!="     { TUNEQUAL _}
  "|"     { TVERT _}
  "where"     { TWHERE _}
  "exit"     { TEXIT _ }
  "Nonces"     { TNONCES _ }
  "return"     { TRETURN _ }
  "Database"     {TDATABSE _}  
  "else"     {TELSE _}
  "="    {TEQUAL _}
  "Formats"     {TFORMATS _}
  "let"     {TLET _}
  "Macros"     {TMACROS _}
  "Main"     {TMAIN p}
  "Mappings"     {TMAPPINGS _}
  "-"     {TMINUS _} 
  "of"     {TOF _}
  "on"    {TON _}
  "or"    {TOR _}
  "Other"     {TOTHER _}
  "+"     {TPLUS _}
  "weakly"    {TWEAKLY _}
---
--  "_" {TUNDERSCORE _}
  "@" {TAT _}  
--  "SeqNumber"        {TSEQNUMBER _}
--  "Untyped"            {TUNTYPED _}
---- Compositional Extension
    "Results"  {TRESULTS _}
    "Private"  {TPRIVATE _}
-- function signature    
    "funsign"  {TFUNSIGN _}
 
%%

protocol :: {Protocol}
  : nameSection              --"Protocol" ":" ident -- opt(NameSection)
    typesSection             --"Types" ":" typedec
    mappingSection           --"Mappings" ":" mappingsdec
    formatsSection           --"Formats" ":" formatsdec 
    knowledgeSection         --"Knowledge" ":" knowdec
    macrosSection            --"Macros" ":" macrosdec 
    actionSection            --"Actions" ":" actionsdec --"Actions" {"(" "Main" ")"      | Nothing} ":" actionsdec --
    goalsSection             --"Goals" ":" goalsdec 
    privateSection           --"Private" ":" privatesdec
    resultsSection           --Results" ":" resultsdec
    abstractionSection       {($1,$2,$3,$4,$5,$6,$7,$8,$9,$10,$11)}

--Name Section it is OPTIONAL
nameSection : {"NoName"}  
			|"Protocol" ":" ident {$3}


{-                        | multComment "Protocol" ":" ident {$4}
                                
-- TODO: Multi-Line comment
multComment: {}
                        | "{#" idents  "#}"              { }  
idents:         ident  {[$1]}         | ident idents {$1:$2}
-}
--Types Section [(Type,[ident])]
typesSection:: {Types} 
        : "Types" ":" typedec {$3}

typedec :: {Types} 
        : type identlist ";" {[($1,$2)]}
        | type identlist ";" typedec {($1,$2):$4}

type :: {Type}
  : ident typeOptions
  {case $1 of
            "Agent" -> Agent False False $2
            "Number" -> Number $2
            "PublicKey" -> PublicKey $2
            "PrivateKey" -> PrivateKey $2
            "SymmetricKey" -> SymmetricKey $2
            "Function" -> Function $2
--            "Nonce"  -> Nonce $2
--            "ImpData" -> ImpData $2
            "Msg" -> Msg $2
            "Untyped" -> Untyped $2 --I excluded this for a period, must be working in the last version
            "SeqNumber" -> SeqNumber $2 --I excluded this for a period, must be working in the last version
            "Bool" -> Boolean $2
            _ -> error ("Unknown type "++$1)
            }

funsign :: {([Type],Type)}
          : type "->" type {([$1],$3)}
          | typelist "->" type {($1,$3)}
                       
typeOptions :: {TypeOpts}
    : {[]} 
    | "[" typeOption "]" {$2}
    
typeOption :: {TypeOpts}
    : ident "=" ident {[Option ($1,Atom $3)]}
    | ident "=" ident "," typeOption {(Option ($1,Atom $3)):$5}
-- uncomment here to enable function type signature
--    | funsign {[FunSign $1]}
--    | funsign "," typeOption {(FunSign $1):$3}

identlist :: {[Ident]}
        : ident {[$1]}
        | ident "," identlist {$1:$3}
        
-- Mappings Section --{[(Ident,[Type],Type)]}    , it is OPTIONAL
mappingSection :: {Mappings}
    : {[]}
    | "Mappings" ":" mappingsdec {$3}
mappingsdec   :: {Mappings}--{[(Ident,[Type],Type)]} --Type??
    : ident ":" typelist "->" type ";" mappingsdec {(($1, $3,$5):$7)} 
    | ident ":" typelist "->" type ";"  {([($1, $3,$5)])} 
    
typelist :: {[Type]} 
        : type {[$1]}
        | type "," typelist {$1:$3}
        
--Formats Section, it is OPTIONAL
formatsSection:: {Formats}
        : {[]}
        | "Formats" ":" formatsdec {$3}

formatsdec  ::{Formats}
    : ident "(" typelist ")" ";" formatsdec {(($1, $3): $6)}
    | ident "(" typelist ")" ";" {([($1, $3)])}

--KNOWLEDGE SECTION 
knowledgeSection :: {Knowledge}
        : "Knowledge" ":" knowdec {$3}

knowdec :: {Knowledge}
: knowspec "where" wheredec {($1,$3)}
| knowspec {($1,[])}

knowspec :: {[(Ident,[Msg])]}
  : ident ":" msglist ";" knowspec {(($1,$3):$5)}
  | ident ":" msglist ";" {[($1,$3)]} 

wheredec :: {[(Msg,Msg)]}
: msg "!=" msg ";" {[($1,$3)]}
| msg "!=" msg "," wheredec {(($1,$3):$5)}

msglist :: {[Msg]}
  : msgNOP "," msglist {$1:$3}
  | msgNOP {[$1]}

msg :: {Msg}
 : msgNOP                {$1}
--| msg "," msg            {[$1,$3]} --(1)-- this is an option to replace Cat or (2)
--  | msg "," msg            {Comp Cat [$1,$3]}  -- original
-- | msglist --(2)-- this may replace Cat, or  (3)

msgNOP :: {Msg}
      : ident                            {Atom $1}
-------------------------------------------------------------------------------------------
--    | "{" msg "}" msgNOP                   {Comp Crypt [$4,$2]}     --original
--    | "{|" msg "|}" msgNOP                 {Comp Scrypt [$4,$2]}     --original  
--  | "{" msglist "}" msgNOP               {Comp Crypt ($4:$2)}     -- 9/5
--    | "{|" msglist "|}" msgNOP             {Comp Scrypt ($4:$2)}        --9/5
-------------------------------------------------------------------------------------------
--     | "{" msg "}" msgNOP                   {signOrCrypt $4 $2} -- Working 12.6
     | "{" msglist "}" msgNOP                   {signOrCrypt $4 $2}
--    | "{|" msg "|}" msgNOP                 {Comp Scrypt [$4,$2]} -- Working 12.6
     | "{|" msglist "|}" msgNOP                 {Comp Scrypt [] ($4:$2)} 
       | ident typeOptions "(" msglist ")"    {        -- CASE WITH TYPE OPTIONS
                                            if      $1=="inv"     then     (if ((length $4)== 1) then Comp Inv [] $4 else error("inv takes one argument") ) 
                                            else if $1=="xor"     then     (if ((length $4)== 2) then Comp Xor [] $4 else error("xor takes two arguments"))
                                            else if $1=="exp"     then     (if ((length $4)== 2) then Comp Exp [] $4 else error("exp takes two arguments") )
                                            else if $1=="hash"     then     Comp Hash $2 $4
                                            --(if ((length $4)== 1) then  Comp Hash $4 else error("Hash takes one argument"))
                                            else if $1=="mac"     then     Comp Mac $2 $4
                                            --(if ((length $4)== 2) then Comp Mac $4 )
                                            else if $1=="mult"     then     (if ((length $4)== 2) then Comp Mult [] $4 else error("mult takes two arguments"))
                                            else if $1=="pk"     then     (if ((length $4)== 1) then Comp Pk [] $4 else error("pk takes one argument"))
                                            else if $1=="sk"     then     (if ((length $4)== 1) then Comp Sk [] $4 else error("sk takes one argument"))
                                            -- else if $1=="scrypt" then (if ((length $3)== 2) then Comp Scrypt $3 else error("exp takes two arguments"))
                                            else if $1=="scrypt" then (if ((length $4)== 2) then Comp Scrypt $2 $4 else error("scrypt takes two arguments, use formats!"))
                                            else if $1=="crypt" then (if ((length $4)== 2) then Comp Crypt $2 $4 else error("crypt takes two arguments, use formats"))
                                            else if $1=="shk"     then (if ((length $4)== 2) then Comp Shk [] $4 else error("exp takes two arguments"))
                                            else if $1=="sign"     then (if ((length $4)== 2) then Comp Sign $2 $4 else error("Sign takes two arguments"))
                                            else if $1=="signh" then (if ((length $4)== 2) then Comp SignH $2 $4 else error("Signh takes two arguments"))
                                            else if $1=="fsignh" then (if ((length $4)== 2) then Comp FsignH $2 $4 else error("Fsignh takes two arguments"))
                                                                                        else if $1=="pvk"     then     (if ((length $4)== 1) then Comp PvK [] $4 else error("pvk takes one argument"))
                                                                                        else if $1=="pbk"     then     (if ((length $4)== 1) then Comp PbK [] $4 else error("pbk takes one argument"))
                                                                                        else if $1=="pub"     then     (if ((length $4)== 1) then Comp Pub [] $4 else error("pub takes one argument"))
                                            -- else if (not (( elem $1 (map fst f)) || ( elem $1  (map fst m)) )) then error ("APS Parse error: " ++ f ++ " undeclared identifier\n" )
                                            -- || ( elem $1 (map fst c))
                                            else case $4 of
                                             [x] -> Comp Apply [] ((Atom $1):[x])
                                             _ -> Comp Apply [] ((Atom $1):$4)} -- Working 12.6 (3) -- this may replace Cat, or something else?!
                                             -- _ -> Comp Apply ((Atom $1):[Comp Cat $3])} -- original    
           | ident "(" msglist ")"            {        -- CASE NO TYPE OPTIONS
                                            if         $1=="inv"     then     (if ((length $3)== 1) then Comp Inv [] $3 else error("inv takes one argument") ) 
                                            else if $1=="xor"     then     (if ((length $3)== 2) then Comp Xor [] $3 else error("xor takes two arguments"))
                                            else if $1=="exp"     then     (if ((length $3)== 2) then Comp Exp [] $3 else error("exp takes two arguments") )
                                            else if $1=="hash"     then     Comp Hash [] $3
                                            --(if ((length $3)== 1) then  Comp Hash $3 else error("Hash takes  one argument"))
                                            else if $1=="mac"     then     Comp Mac [] $3
                                            --(if ((length $3)== 2) then Comp Mac $3 )
                                            else if $1=="mult"     then     (if ((length $3)== 2) then Comp Mult [] $3 else error("mult takes two arguments"))
                                            else if $1=="pk"     then     (if ((length $3)== 1) then Comp Pk [] $3 else error("pk takes one argument"))
                                            else if $1=="sk"     then     (if ((length $3)== 1) then Comp Sk [] $3 else error("sk takes one argument"))
                                            -- else if $1=="scrypt" then (if ((length $3)== 2) then Comp Scrypt $3 else error("exp takes two arguments"))
                                            else if $1=="scrypt" then (if ((length $3)== 2) then Comp Scrypt [] $3 else error("scrypt takes two arguments, use formats!"))
                                            else if $1=="crypt" then (if ((length $3)== 2) then Comp Crypt [] $3 else error("crypt takes two arguments, use formats"))
                                            else if $1=="shk"     then (if ((length $3)== 2) then Comp Shk [] $3 else error("exp takes two arguments"))
                                            else if $1=="sign"     then (if ((length $3)== 2) then Comp Sign [] $3 else error("Sign takes two arguments"))
                                            else if $1=="signh" then (if ((length $3)== 2) then Comp SignH [] $3 else error("Signh takes two arguments"))
                                            else if $1=="fsignh" then (if ((length $3)== 2) then Comp FsignH [] $3 else error("fsignh takes two arguments"))
                                                                                        else if $1=="pvk"     then     (if ((length $3)== 1) then Comp PvK[] $3 else error("pvk takes one argument"))
                                                                                        else if $1=="pbk"     then     (if ((length $3)== 1) then Comp PbK [] $3 else error("pbk takes one argument"))
                                                                                        else if $1=="pub"     then     (if ((length $3)== 1) then Comp Pub [] $3 else error("pub takes one argument"))
                                            -- else if (not (( elem $1 (map fst f)) || ( elem $1  (map fst m)) )) then error ("APS Parse error: " ++ f ++ " undeclared identifier\n" )
                                            -- || ( elem $1 (map fst c))
                                            else case $3 of
                                             [x] -> Comp Apply [] ((Atom $1):[x])
                                             _ -> Comp Apply [] ((Atom $1):$3)} -- Working 12.6 (3) -- this may replace Cat, or something else?!
                                             -- _ -> Comp Apply ((Atom $1):[Comp Cat $3])} -- original    
      | "("msg")"            {$2}
--      | "(" msglist ")"        {Comp Cat $2}    -- new 
-- 


--MACROS SECTION {(Ident, [Ident], Msg)}

macrosSection ::{Macros}
        : {[]}
        | "Macros" ":" macrosdec {$3}

macrosdec   :: {Macros} 
    : ident "(" identlist ")" "=" msg ";" macrosdec {(($1,$3,$6) : $8)}--($1:macrosFormatsMappings)}
    | ident "(" identlist ")" "=" msg ";" {([($1,$3,$6)])}--($1:macrosFormatsMappings)}
{-
macrosSection ::{Macros}
        : {[]}
        | "Macros" ":" macrosdec {$3}

macrosdec   :: {Macros} 
    : ident "(" identlist ")" "=" msg ";" macrosdec {((Atom $1,(map Atom $3),$6) : $8)}--($1:macrosFormatsMappings)}
    | ident "(" identlist ")" "=" msg ";" {([(Atom $1,(map Atom $3),$6)])}--($1:macrosFormatsMappings)}
-}
    

-- ACTIONS SECTION
actionSection :: {Actions}
        : "Actions" ":" actionsdec {$3}
        | "Actions" "(" "Main" ")" ":" actionsdec {$6}
    
actionsdec :: {Actions}
  : action {[$1]}
  | action actionsdec {($1:$2)}

action :: {Action}
  : channel ":" msg "%" msg "!" msg  {Exchange ($1,$3,Just $5,Just $7)}
  | channel ":" msg "%" msg     {Exchange ($1,$3,Just $5,Nothing)}
  | channel ":" msg      {Exchange ($1,$3,Nothing,Nothing)}
  | peer ":" type identlist {CreateFresh (($1,DecFresh,$1),($3,$4))} -- allowing fresh declaration e.g., A: Nonce X  
  | "if" "(" conditiondec ")" "then" action "else" action "fi" {If ($3, $6 ,$8)}
  | "let" ident "=" msg { Let ($2, $4)}
  | action "or" action {Or ($1,$3)}
  
conditiondec
    : ident {$1}

channeltype :: {ChannelType}
  : "*->*" {Secure}
  | "*->" {Authentic}
  | "*->>" {FreshAuthentic}
  | "*->>*" {FreshSecure}
  | "->*" {Confidential}
  | "->" {Insecure}

channeltypeG :: {ChannelType}
  : "*->*" {Secure}
  | "*->>*" {FreshSecure}
  | "*->" {Authentic}
  | "*->>" {FreshAuthentic}
  | "->*" {Confidential}
 -- | "->>*" {FreshConfidential}  ?? to add in Lexer
  | "->" {Insecure}

channel :: {Channel} 
  : peer channeltype peer {($1,$2,$3)} 
 
channelG :: {Channel} 
  : peer channeltypeG peer {($1, $2,$3)} 
 
 
peer :: {Peer}
  : ident {($1, Atom $1,False,Nothing)}
  | "[" ident "]" {($2, Atom $2, True,Nothing)}
  | "[" ident ":" msg "]" {($2, Atom $2, True, Just $4)}

  
--GOALS SECTION
goalsSection :: {Goals}
        : "Goals" ":" goalsdec {$3}
goalsdec :: {Goals}
  : goal {[$1]}
  | goal goalsdec {$1:$2}

goal :: {Goal}
  : channelG ":" msg      
  {(ChGoal $1 $3)}
  | peer "authenticates" peer "on" msg 
  --- {(Authentication $1 $3 $5)}
  {(ChGoal ($3,FreshAuthentic,$1) $5)}
  | peer "weakly" "authenticates" peer "on" msg 
  {(ChGoal ($4,Authentic,$1) $6)}
  --- {(WAuthentication $1 $4 $6)}
  | msg "secret" "between" peers 
  {(Secret $1 $4)}
  | msg "secret" "of" peers 
  {(Secret $1 $4)}

peers :: {[Peer]}
peers : peer {[$1]}
| peer "," peers {($1:$3)}

abstractionSection :: {Abstraction}
        : {[]}
        | "Abstraction" ":" abslist {$3}
abslist :: {Abstraction}
        : ident "->" msgNOP {[($1,$3)]}
        | ident "->" msgNOP ";" abslist {(($1,$3):$5)}
----- Compositional Extension
privateSection:: {Private}
        : {[]}
        | "Private" ":" msglist {$3}
-- privatesdec:: {[Msg]} : ident ";" {[$1]}        | ident "," privatesdec {$1:$3}
resultsSection:: {Results}
        : {[]}
        | "Results" ":" resultsdec {$3}
resultsdec::{Results}
        : ident "," ident ":" msgNOP ";" {[($1,$3,$5)]}
        | ident "," ident ":" msgNOP ";" resultsdec {($1,$3,$5):$7}
---------------------------------

{


happyError :: [Token] -> a
happyError tks = error ("APS Parse error at " ++ lcn ++ "\n" )
    where
    lcn =     case tks of
          [] -> "end of file"
          tk:_ -> "line " ++ show l ++ ", column " ++ show c ++ " - Token: " ++ show tk
            where
            AlexPn _ l c = token_posn tk
}
