{-
SPS- Security Protocol Specification language 
Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2015

APCC - Automated Protocol Composition Checker

Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2014-2015
Copyright Paolo Modesti, School of Computing Science, Newcastle University 2015

APS- Authentication Protocol Specification language 
Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2013-2015

This software is released under the BSD License (see BSD_License.txt)


-}

module Main where
import Lexer
import Parser
import Ast
import Msg
import StrandAST
import Preprocessor
import GeneratePlainStrands
import OperationalStrands
import ToJava
import ToJS
import ToFormat
import ToIF
import ToAnB                   -- needs a theory file if formats are used
import ToAnBf
import ToTheory
import ToPv                    -- The Hack of exp (fixing g) -- import ToPiFinal -- The way it is supposed to be!
import System.Environment
import Data.Hashable
import Data.List
import TypeSystem_OS
import StrandMP
import System.Directory(getDirectoryContents)
import qualified Data.List as List
---------------------------------------------------------------------
---------------------------------------------------------------------
helpStr="SPS/APS Security Protocol Specification language  2013-2015\n"++
		"APCC Automated Protocol Composition Checker 2015\n\n"++
		"APS --{opt} <filename>\n {opt} must be one of the following (case insensitive):\n"++
            "  --help\t Generates this list\n"++
		--	"  --java\t Generates Java\n"++ not Tested, so exclude it!
			"  --js\t\t Generates JavaScript\n"++
			"  --pv\t\t Generates Applied-pi\n"++
            "  --oos\t\t Generates Operational Strands (all details)\n"++
			"  --los\t\t Generates Operational Strands (readable version)\n"++   
            "  --tos\t\t Generates Type-checked Operational Strands (all details)\n"++
		--	"  --pro\t\t Generates the Protocol\n"++ -- Internal business!
            "  --str\t\t Generates Plain Strands\n"++
			"  --fr\t\t Generates Format Classes in Java (skeleton)\n"++
		--	"  --anb\t\t Generates AnB code\n"++ -- Issues with the OFMC, long names, 
			"  --if\t\t Generates IF code\n"++
			"  --anbf\t Generates AnB code (No Formats)\n"++
        --    "  --thy\t\t Generates AnB  theory file\n" ++ -- Issues with the OFMC, long names
            "  --smp\t\t Generates SMP (sub-message patterns)\n\n\n" ++ 
            "APS  --pcc <foldername>\t Checks if the prococols in the folder are parellel composable\n"++
            "\n ------------------------------------\n Have Fun\n Kindly report any problem to Omar Almousa at omal@dtu.dk"
---------------------------------------------------------------------
---------------------------------------------------------------------
allTogetherJava x = toJavas (opStrands (generateStrands x))
allTogetherJS x = toJSs (opStrands (generateStrands x))
allTogetherPi3 x = toPIsFinal (getWheres x) (getAgents x)  (opStrands (generateStrands x))
allTogetherIf x frmats= toIFs  frmats (excOrdProt x) (opStrands (generateStrands x))
allTogetherLOS x = concatMap strndPP(opStrands (generateStrands x)) 
allTogetherOOS x = concatMap show(opStrands (generateStrands x)) 
--------------------------------------------------------------
fullClean :: Protocol -> Protocol
fullClean prot =  getEmbdFresh . chkNoFrshAll . mcrosLetsFrmtsMppngs . protprivate2goals $ prot

main =  do 
            args <- getArgs
            if null args then putStrLn helpStr 
                else do
                    output <- return (head args)
                    let lwo = lowerize output
                    case lwo of
                        "--pcc"-> compchecker (tail args)
                        _ -> apscompiler (last args) output

apscompiler f opt = do
    aps <- readFile f
    let 
        tokns = alexScanTokens aps
        prot = apspareser tokns
        cleanProt = fullClean prot
        fileName = take (length f - 4) f ++ "X"
        extension = reverse (take 4 (reverse f))
        (code,ext) = if extension /= ".aps" then error "This is the APS compiler; your file extension must be .aps\n" 
                        else case opt of
                             --   "--java" -> (allTogetherJava cleanProt, "_prog.java")
                                "--js"   -> (allTogetherJS cleanProt, "_prog.js")
                                "--fr"   -> (concat $ formClasses $ getFormats cleanProt, ".java") --or Form
                                "--if"   -> (allTogetherIf cleanProt (getFormats cleanProt), ".if")
                                "--pv"   -> (allTogetherPi3 cleanProt, ".pv") -- with exp hack (fixing g)
                                "--los"  -> (allTogetherLOS cleanProt, ".los")
                                "--oos"  -> (allTogetherOOS cleanProt, ".oos")
                                "--str"  -> (show(generateStrands cleanProt), "_Op.str")
                                "--tst"  -> (show(generateStrands cleanProt), ".tst")
                                "--pro"  -> (show cleanProt, ".pro")
                             --   "--anb"  -> (anbs cleanProt, ".AnB")
                                "--anbf" -> (anbsf cleanProt, "F.AnB")
                             --   "--thy"  -> (thy (getFormats cleanProt) ++ stdTheory, ".thy")
                                "--ctx"  -> (show (buildContext cleanProt),".ctx")
                                "--tos"  -> (ttsOOS cleanProt,".tos")
                                "--smp"  -> (showttsSMP cleanProt,".smp")
                                _        -> error("Unrecognised output format, please use:\n"++helpStr)
    case ext of 
           ".tst" -> print (hash code) 
           "help" -> print helpStr
           _      -> writeFile (fileName ++ ext) code
                  
compchecker xs = if null xs then putStrLn "No folder specified - A folder is required for --pcc option"
                     else do
                        putStrLn ("Analysing protocols in folder: " ++ show xs)
                        dir <- getDirectoryContents (head xs) >>= return . filter (List.isSuffixOf ".aps") -- added afterwards
                        file <- mapM (readFile . (\ x -> head xs ++ "/" ++ x)) dir -- also!                
                        let protocols = map (apspareser . alexScanTokens) file
                        let protocols1 = map fullClean protocols
                        putStrLn "Composition checker"
                        putStrLn "Are protocols Parallel Composable?"
                        let unifications = nub . smpDisjointCheckerU $ protocols1
                        if null unifications then putStrLn "YES!" 
                                             else do 
                                                     putStrLn ("NO!\nThere are at least " ++ show (length unifications) ++ " possible unifications:\n" ++ (showuniPairs . countUni $ unifications) ++ "\n\n") 
                                                     putStrLn (showUni unifications)                       

-- pairs of protocol names unifications
uniPairs :: Unifications -> [(String,String)]
uniPairs s = [(p1,p2) | ((_,p1),(_,p2)) <- s ]

-- count the occurrences of unificatiosn
countUni ::  Unifications -> [((String,String),Int)]
countUni s =  sortBy (\(_,n1) (_,n2) -> if n1 > n2 then LT else GT) . map (\x -> (head x, length x)) . groupBy (\(p1,p2) (p3,p4) -> p1==p3 && p2==p4) . sort . uniPairs $ s

showuniPairs :: [((String,String),Int)] -> String
showuniPairs [] = ""
showuniPairs (((p1,p2),n):xs) = p1 ++ " - " ++ p2 ++ ": " ++ show n ++ "\n" ++ showuniPairs xs 
