{-

SPS/APS - Security Protocol Specification langauge
Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2013-2015

All Rights Reserved.

-}
module ToJava where
import Analyze
import ComposeM
import StrandAST
import Msg
import OperationalStrands
import GeneratePlainStrands
--import JSJavaUtility
import Data.List
import Data.List.Split

------------
-- toJavasFiles x = splitOn "#" (toJavas x)
--- Best version for putStr, delete others
toJavas []= ""
toJavas  (x:xs)=   (toJava x) ++ (toJavas xs)

-- toJavas  types (x:xs)=   (toJava types protx) ++ (toJavas types xs)
------------------

------------------

-- Yet an older other version suitable for print and producing  two strings that we can write later to files.java
-- toJavas (x:xs)=  (splitOn "//###" (cleanString (toJava x))) : (toJavas xs)
-- toJavas (x:xs)=  (cleanString (toJava x)) ++ (toJavas xs)
--splitOn ";"
-------------
-- toJava:: [Types] -> Strand -> String
-- toJava::  Strand -> String
toJava s@(Strand n o know rest)= (header n o know) ++ (toJava0 [] rest)
------------------
--- checked is to eliminate duplicated and redundant checks produced from calling analyze several time!
-- [Strand0] ->Strand0-> String
toJava0 checked (Snd det m rest) = "ch.send(" ++ (getIdent m)++ ");\n"++ (toJava0 checked rest)
toJava0 checked (Rcv det m rest) = "byte[] " ++ (getIdent m) ++ " = ch.receive();\n" ++ (toJava0 checked rest)
toJava0 checked s@(V Check m1 m2 rest)     =  if (elem ("Check",m1,m2) checked || elem ("Check",m2,m1) checked) 
                                            then toJava0 checked rest 
                                            else "if(" ++ (getIdent m1)++ " != " ++ (getIdent m2) ++ ") error();\n" ++ 
                                                        (toJava0 (("Check",m1,m2):checked) rest)
toJava0 checked s@(V ver m1 m2 rest)     =  toJava0 checked rest
-- a is HERE
toJava0 checked (How (Get n) m1 m2 m3 rest) = "byte[] "++ (getIdent m1) ++ " = " ++ (getIdent m3) ++ "a.get(" ++ show(n) ++ ");\n" ++
                                                (toJava0 checked rest)
toJava0 checked (How (Project n) m1 m2 m3 rest) = "byte[] "++ (getIdent m1) ++ " = proj" ++ show(n)++ "("  ++ (getIdent m2) ++
                                                    ");\n" ++ (toJava0 checked rest)
toJava0 checked (How des m1 m2 m3 rest)=  if(elem ("Open",m2,m3) checked) then toJava0 checked rest 
                                            else "byte[] "++ (getIdent m1) ++ " = " ++ (lowerize (show(des)))++"(" ++ (getIdent m2)++ ", " ++ 
                                                (getIdent m3)++");\n" ++ (toJava0 (("Open",m2,m3):checked) rest)
-- a is HERE
toJava0 checked (Create m1 m2 m3 rest)= (getIdent m3) ++" " ++(getIdent m1) ++ "a = new "++ (getIdent m3)++ "("++(getIdent m2)++
                                                ");\n" ++(toJava0 checked rest)
toJava0 checked (Add event id m ps rest)= "" ++ (toJava0 checked rest)
toJava0 checked (Fresh typ ms orms rest)= "//FRESH VALUE IS CREATED!! To Be handled later!\n" ++ (toJava0 checked rest)
toJava0 checked (EOS _)                 =  "}\n //#"
{-
toJava0 checked (Bind m1 m2 rest)=if (elem (m1,m2) checked) then toJava0 checked rest
                                                -- else "byte[] "++ (getIdent m1)++ " = " ++ (getIdent m2) ++";\n"++ (toJava0 ((m1,m2):checked) rest)
-- toJava0 checked s@(Check m1 m2 rest)     =  if (elem (m1,m2) checked) then toJava0 checked rest
                                                -- else "if(" ++ (getIdent m1)++ " != " ++ (getIdent m2) 
                                        -- ++ ") throwException(" ++ (getIdent m1)++ ");\n" ++ (toJava0 ((m1,m2):checked) rest)
-- toJava0 checked (How (Open key) m1 m2 rest)=  "byte[] "++ (getIdent m1) ++ " = Open (" ++ (getIdent key)++ "," ++ (getIdent m2) ++");\n" ++ (toJava0 checked  rest)
--another way to deal with gets:
-- toJava0 (How (Get n for) m1 m2 rest)     = (checked, ("byte[] "++ (getIdent m1) ++ " = Get(" ++ show(n) ++ ", " ++ (getIdent for) ++ ", " ++ (getIdent m2) ++ ");\n"  ++ (toJava0 checked rest)))
-}
---------------
header n o k= "//Auto-Generated Code by APS2Java\n\n\n//Agent  "++ o ++" function in protocol:"++ n ++"  is  \nvoid prog_"++ o ++"_"++n++"("++ (paramz k) ++", Channel ch){\n" 
-- ++ "\nChannel ch= new Channel();\n"
params []  = ""
params [(a,b)] = (typeOf b) ++ " "++ (getIdent a)
params ((a,b):y:z) =(typeOf b) ++ " " ++ (getIdent a) ++"," ++ (params(y:z))
paramz a= params  a
--------------------------
typeOf m= "byte[]"-- to be handled later
---------------
cleanString st = [ c | c <- st, c /= '/' && c/= '"' ]
------------------
-- getIdent (Atom ('f':'r':'m':rst))= capFst rst
-- getIdent (Atom ('m':'p':'p':rst))= lowerize rst
-- getIdent (Atom x)= lowerize x
-- getIdent (Comp Apply ((Atom ('f':'r':'m':rst)):xs))= "new "++ (capFst rst) ++ "(" ++ (getIdents xs)++").encode()"
-- getIdent (Comp Apply (x:xs))= (getIdent x) ++ "(" ++ (getIdents xs)++")"
-- getIdent (Comp Cat ls)= "(" ++ (getIdents ls)++")"
-- getIdent (Comp op ls)= (lowerize (show(op))) ++ "(" ++ (getIdents ls)++")"

-----------
-- getIdents []=" "
-- getIdents [x]= getIdent x
-- getIdents (x:y:z)= (getIdent x) ++ ", " ++ (getIdents(y:z))
----------------